package com.oracle.jpa.bugtest;

import jakarta.persistence.EntityManager;
import jakarta.persistence.EntityManagerFactory;
import jakarta.persistence.EntityTransaction;
import jakarta.persistence.Persistence;
import jakarta.persistence.Query;
import org.junit.jupiter.api.Test;

import java.lang.reflect.Field;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;

public class TestJPQLUPDATEWithEmbeddables {

    private static final String PU_NAME = "test-jpa-pu-derby";

    @Test
    public void testJPQLUpdateWithEmbeddablesWithNull() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory(PU_NAME);
        EntityManager em = emf.createEntityManager();
        EntityTransaction tx = em.getTransaction();

        try {
            tx.begin();
            Query query = em.createQuery("UPDATE House o SET o.garage=?2 WHERE (o.parcelId=?1)");
            query.setParameter(1, "TestEmbeddable-304-3655-30");
            query.setParameter(2, null);
            query.executeUpdate();
            tx.commit();
            House house = em.find(House.class, "TestEmbeddable-304-3655-30");
            assertNotNull(house);
            assertNull(house.garage);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        }
        finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
            emf.close();
        }
    }

    @Test
    public void testJPQLUpdateWithEmbeddablesWithValue() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory(PU_NAME);
        EntityManager em = emf.createEntityManager();
        EntityTransaction tx = em.getTransaction();

        try {
            Garage garage = new Garage(100, "Detached");
            tx.begin();
            Query query = em.createQuery("UPDATE House o SET o.garage=?2 WHERE (o.parcelId=?1)");
            query.setParameter(1, "TestEmbeddable-304-3655-30");
            query.setParameter(2, garage);
            query.executeUpdate();
            tx.commit();
            House house = em.find(House.class, "TestEmbeddable-304-3655-30");
            assertNotNull(house);
            assertEquals(garage, house.garage);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        }
        finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
            emf.close();
        }
    }

    @Test
    public void testJPQLUpdateWithEmbeddablesWithValue02() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory(PU_NAME);
        EntityManager em = emf.createEntityManager();
        EntityTransaction tx = em.getTransaction();

        try {
            Garage garage = new Garage(100, "Detached");
            tx.begin();
            Query query = em.createQuery("UPDATE House o SET o.garage.area=?2, o.garage.type=?3 WHERE (o.parcelId=?1)");
            query.setParameter(1, "TestEmbeddable-304-3655-30");
            query.setParameter(2, garage.area);
            query.setParameter(3, garage.type);
            query.executeUpdate();
            tx.commit();
            House house = em.find(House.class, "TestEmbeddable-304-3655-30");
            assertNotNull(house);
            assertEquals(garage, house.garage);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        }
        finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
            emf.close();
        }
    }

    @Test
    public void testUpdateWithEmbeddablesWithNull() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory(PU_NAME);
        EntityManager em = emf.createEntityManager();
        EntityTransaction tx = em.getTransaction();

        try {
            tx.begin();
            House house = em.find(House.class, "TestEmbeddable-304-3655-30");
            house.garage = null;
            em.merge(house);
            tx.commit();
            house = em.find(House.class, "TestEmbeddable-304-3655-30");
            assertNotNull(house);
            assertNull(house.garage);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        }
        finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
            emf.close();
        }
    }

    @Test
    public void testUpdateWithEmbeddablesWithValue() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory(PU_NAME);
        EntityManager em = emf.createEntityManager();
        EntityTransaction tx = em.getTransaction();

        try {
            Garage garage = new Garage(100, "Detached");
            tx.begin();
            House house = em.find(House.class, "TestEmbeddable-304-3655-30");
            house.garage = garage;
            em.merge(house);
            tx.commit();
            house = em.find(House.class, "TestEmbeddable-304-3655-30");
            assertNotNull(house);
            assertEquals(garage, house.garage);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        }
        finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
            emf.close();
        }
    }

    @Test
    public void testJPQLSelectWithEmbeddablesWithValue() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory(PU_NAME);
        EntityManager em = emf.createEntityManager();
        EntityTransaction tx = em.getTransaction();

        try {
            Garage garage = new Garage(200, "Attached");
            Query query = em.createQuery("SELECT o FROM House o WHERE o.garage=?1", House.class);
            query.setParameter(1, garage);
            House house = (House) query.getSingleResult();
            assertNotNull(house);
            assertEquals(garage, house.garage);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        }
        finally {
            em.close();
            emf.close();
        }
    }


    @Test
    public void testGarageField01() throws NoSuchFieldException, IllegalAccessException {
        Garage anObject = new Garage(100, "Detached");
        Field attributeField = House.class.getDeclaredField("garage");
        Object result = attributeField.get(anObject);
        System.out.println(result);
    }

    @Test
    public void testGarageField02() throws NoSuchFieldException, IllegalAccessException {
        Garage garage = new Garage(100, "Detached");
        House anObject = new House("TestEmbeddable-304-3655-30", garage);
        Field attributeField = House.class.getDeclaredField("garage");
        Object result = attributeField.get(anObject);
        System.out.println(result);
    }
}
