package com.oracle.jpa.bugtest;

import java.lang.reflect.Constructor;
import java.lang.reflect.RecordComponent;
import java.time.Instant;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.stream.Collectors;

import jakarta.persistence.EntityManager;
import jakarta.persistence.EntityManagerFactory;
import jakarta.persistence.Persistence;
import org.eclipse.persistence.descriptors.ClassDescriptor;
import org.eclipse.persistence.descriptors.copying.InstantiationCopyPolicy;
import org.eclipse.persistence.descriptors.copying.RecordCopyPolicy;
import org.eclipse.persistence.internal.descriptors.ObjectBuilder;
import org.eclipse.persistence.internal.descriptors.RecordInstantiationPolicy;
import org.eclipse.persistence.internal.helper.DatabaseField;
import org.eclipse.persistence.internal.sessions.AbstractRecord;
import org.eclipse.persistence.internal.sessions.AbstractSession;
import org.eclipse.persistence.internal.sessions.ArrayRecord;
import org.eclipse.persistence.jpa.JpaEntityManager;
import org.eclipse.persistence.mappings.AggregateObjectMapping;
import org.eclipse.persistence.mappings.DatabaseMapping;
import org.eclipse.persistence.mappings.DirectToFieldMapping;
import org.eclipse.persistence.queries.ReadObjectQuery;
import org.eclipse.persistence.sessions.Session;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;

public class TestRecord {

    private static final String PU_NAME = "test-jpa-record-pu-derby";

    private final int ID_INT = 1;
    private final String ID_STRING = "a";
    private final float ID_FLOAT = 1.1F;
    private final Instant ID_INSTANT = Instant.parse("2024-05-27T10:15:30.00Z");

    private final String NAME1 = "abcde";
    private final String NAME2 = "ijkl";
    private final String NAME3 = "xyz";

    @Test
    public void persistTest() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        assertNotNull(entityManagerFactory);

        EntityManager em = entityManagerFactory.createEntityManager();

        try {
            em.getTransaction().begin();

            em.createNativeQuery("DELETE FROM TEST_TAB_RECORD").executeUpdate();

            RecordEntity entity = new RecordEntity();
            entity.setId(new RecordPK(ID_INT, ID_STRING));
            entity.setRecordAttribute(new RecordAttribute(new RecordNestedAttribute(NAME1, NAME2), NAME3));
            em.persist(entity);
            em.getTransaction().commit();
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        } finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
        }
    }

    @Test
    public void findTest() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        assertNotNull(entityManagerFactory);

        EntityManager em = entityManagerFactory.createEntityManager();

        try {
            RecordEntity entity = em.find(RecordEntity.class, new RecordPK(ID_INT, ID_STRING));
            assertNotNull(entity);

            assertEquals(ID_INT, entity.getId().id_int());
            assertEquals(NAME1, entity.getRecordAttribute().nestedAttribute().name1());
            assertEquals(NAME2, entity.getRecordAttribute().nestedAttribute().name2());
            assertEquals(NAME3, entity.getRecordAttribute().name3());
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        } finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
        }
    }

    @Test
    public void queryTest() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        assertNotNull(entityManagerFactory);

        EntityManager em = entityManagerFactory.createEntityManager();

        try {
            List<RecordEntity> result = em
                    .createQuery("SELECT OBJECT(o) FROM RecordEntity o WHERE o.recordAttribute = :recordAttribute")
                    .setParameter("recordAttribute",
                            new RecordAttribute(new RecordNestedAttribute(NAME1, NAME2), NAME3))
                    .getResultList();

            for (RecordEntity entity : result) {
                assertNotNull(entity);

                assertEquals(ID_INT, entity.getId().id_int());
                assertEquals(NAME1, entity.getRecordAttribute().nestedAttribute().name1());
                assertEquals(NAME2, entity.getRecordAttribute().nestedAttribute().name2());
                assertEquals(NAME3, entity.getRecordAttribute().name3());
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        } finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
        }
    }

    @Test
    public void queryNEWTest() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        assertNotNull(entityManagerFactory);

        EntityManager em = entityManagerFactory.createEntityManager();

        try {
            List<RecordPK> result = em.createQuery(
                    "SELECT NEW com.oracle.jpa.bugtest.RecordPK(o.id.id_int, o.id.id_string) FROM RecordEntity o WHERE o.recordAttribute = :recordAttribute")
                    .setParameter("recordAttribute",
                            new RecordAttribute(new RecordNestedAttribute(NAME1, NAME2), NAME3))
                    .getResultList();

            for (RecordPK entity : result) {
                assertNotNull(entity);
                assertEquals(ID_INT, entity.id_int());
                assertEquals(ID_STRING, entity.id_string());
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        } finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
        }
    }

    @Test
    public void testQueryWithRecordResult() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        assertNotNull(entityManagerFactory);

        EntityManager em = entityManagerFactory.createEntityManager();

        try {
            // Simulate the query for Roman numerals less than or equal to 20
            List<RomanNumeral> result = em
                    .createQuery("SELECT r FROM RomanNumeral r WHERE r.value <= :maxValue", RomanNumeral.class)
                    .setParameter("maxValue", 20)
                    .getResultList();

            // Expected results
            List<String> expectedResults = List.of(
                    "eleven XI ( X I )",
                    "five V ( V )",
                    "nineteen XIX ( X I X )",
                    "seven VII ( V I I )",
                    "seventeen XVII ( X V I I )",
                    "thirteen XIII ( X I I I )",
                    "three III ( I I I )",
                    "two II ( I I )");

            // Assert results match expectations
            List<String> actualResults = result.stream()
                    .map(RomanNumeral::toString)
                    .collect(Collectors.toList());

            assertNotNull(actualResults);
            assertEquals(expectedResults.size(), actualResults.size());

            for (int i = 0; i < expectedResults.size(); i++) {
                assertEquals(expectedResults.get(i), actualResults.get(i));
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException();
        } finally {
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }
            em.close();
        }
    }

    @Test
    public void recordInstanceCheck() {
        RecordPK recordPK = new RecordPK(ID_INT, ID_STRING);
        // Object recordPK = new Object();

        if (recordPK instanceof Record) {
            System.out.println("recordPK is Java record");
        }
    }

    @Test
    public void recordCopyBySessionTest() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        Session session = ((JpaEntityManager) entityManagerFactory.createEntityManager()).getActiveSession();
        RecordPK source = new RecordPK(ID_INT, ID_STRING);
        RecordPK clone = (RecordPK) session.copy(source);
        assertEquals(source, clone);
    }

    @Test
    public void cloneNestedRecordTest() {
        NestedDetailRecord nestedDetailRecord = new NestedDetailRecord(ID_FLOAT, ID_INSTANT);
        NestedMasterRecord source = new NestedMasterRecord(ID_INT, ID_STRING, nestedDetailRecord);
        RecordCopyPolicy recordCopyPolicy = new RecordCopyPolicy();
        NestedMasterRecord clone = (NestedMasterRecord) recordCopyPolicy.buildClone(source, null);
        // NestedMasterRecord clone = TestRecord.cloneRecord(source);
        assertEquals(source, clone);
    }

    public static <R extends Record> R cloneRecord(R template) {
        try {
            ArrayList<Class<?>> types = new ArrayList<>();
            ArrayList values = new ArrayList<>();
            for (RecordComponent component : template.getClass().getRecordComponents()) {
                types.add(component.getType());
                Object value = component.getAccessor().invoke(template);
                if (value instanceof Record) {
                    value = cloneRecord((Record) value);
                }
                values.add(value);
            }
            Constructor<? extends Record> canonical = template.getClass()
                    .getDeclaredConstructor(types.toArray(Class[]::new));
            var result = (R) canonical.newInstance(values.toArray(Object[]::new));
            return result;
        } catch (ReflectiveOperationException e) {
            throw new RuntimeException("Record clone failed: " + e, e);
        }
    }

    @Test
    public void buildNewRecordPKInstanceTest() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        Session session = ((JpaEntityManager) entityManagerFactory.createEntityManager()).getActiveSession();
        AbstractRecord databaseRow = prepareRecordPKData((AbstractSession) session);
        Class clazz = RecordPK.class;
        List<DatabaseMapping> mappings = session.getDescriptor(clazz).getMappings();
        RecordAttribute newInstance = (RecordAttribute) buildNewRecordInstance(clazz, mappings, databaseRow,
                (AbstractSession) session);
        assertEquals(NAME1, newInstance.nestedAttribute().name1());
        assertEquals(NAME2, newInstance.name3());
    }

    @Test
    public void buildNewRecordAttributeInstanceTest() {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        Session session = ((JpaEntityManager) entityManagerFactory.createEntityManager()).getActiveSession();
        AbstractRecord databaseRow = prepareRecordAttributeData((AbstractSession) session);
        Class clazz = RecordAttribute.class;
        List<DatabaseMapping> mappings = session.getDescriptor(clazz).getMappings();
        // RecordAttribute newInstance = (RecordAttribute) buildNewRecordInstance(clazz,
        // mappings, databaseRow, (AbstractSession)session);
        ObjectBuilder objectBuilder = new ObjectBuilder(session.getDescriptor(clazz));
        RecordAttribute newInstance = (RecordAttribute) objectBuilder.buildNewRecordInstance(clazz, mappings,
                databaseRow, (AbstractSession) session);
        assertEquals(NAME1, newInstance.nestedAttribute().name1());
        assertEquals(NAME2, newInstance.nestedAttribute().name2());
        assertEquals(NAME3, newInstance.name3());
    }

    public Object buildNewRecordInstance(Class<Record> clazz, List<DatabaseMapping> mappings,
            AbstractRecord databaseRow, AbstractSession session) {
        class TypeValue {
            public TypeValue(Class<?> type) {
                this.type = type;
            }

            Class<?> type;
            Object value;
        }
        Map<String, TypeValue> typeValueMap = new LinkedHashMap<>();
        ReadObjectQuery query = new ReadObjectQuery();
        query.setSession(session);
        for (RecordComponent component : clazz.getRecordComponents()) {
            typeValueMap.put(component.getName(), new TypeValue(component.getType()));
        }
        for (DatabaseMapping mapping : mappings) {
            Object value = null;
            if (mapping instanceof DirectToFieldMapping) {
                value = mapping.valueFromRow(databaseRow, null, query, true);
            }
            if (mapping instanceof AggregateObjectMapping aggregateObjectMapping) {
                ClassDescriptor descriptor = session.getClassDescriptor(aggregateObjectMapping.getReferenceClass());
                value = buildNewRecordInstance((Class<Record>) aggregateObjectMapping.getReferenceClass(),
                        descriptor.getMappings(), databaseRow, session);
            }
            TypeValue typeValue = typeValueMap.get(mapping.getAttributeName());
            typeValue.value = value;
        }
        List values = new ArrayList<>();
        for (TypeValue typeValue : typeValueMap.values()) {
            values.add(typeValue.value);
        }
        return newRecord(clazz, values);
    }

    @Test
    public void newRecordTest() {
        Class<NestedDetailRecord> clazz = NestedDetailRecord.class;
        // Values order is important. It must be same as order of Record attributes.
        List<Object> values = new ArrayList<>();
        values.add(ID_FLOAT);
        values.add(ID_INSTANT);
        RecordInstantiationPolicy<NestedDetailRecord> recordInstantiationPolicy = new RecordInstantiationPolicy<>(
                clazz);
        recordInstantiationPolicy.setValues(values);
        NestedDetailRecord nestedDetailRecord = (NestedDetailRecord) recordInstantiationPolicy.buildNewInstance();
        assertEquals(ID_FLOAT, nestedDetailRecord.floatAttribute());
        assertEquals(ID_INSTANT, nestedDetailRecord.instantAttribute());
    }

    @Test
    public void newRecordWithNullTest() {
        Class<NestedDetailRecord> clazz = NestedDetailRecord.class;
        // Values order is important. It must be same as order of Record attributes.
        List<Object> values = new ArrayList<>();
        values.add(ID_FLOAT);
        values.add(null);
        RecordInstantiationPolicy<NestedDetailRecord> recordInstantiationPolicy = new RecordInstantiationPolicy<>(
                clazz);
        recordInstantiationPolicy.setValues(values);
        NestedDetailRecord nestedDetailRecord = (NestedDetailRecord) recordInstantiationPolicy.buildNewInstance();
        assertEquals(ID_FLOAT, nestedDetailRecord.floatAttribute(), 0);
        assertNull(nestedDetailRecord.instantAttribute());
    }

    public static <R extends Record> R newRecord(Class<R> clazz, List values) {
        return newRecordInternal(clazz, new ArrayList<>(values));
    }

    private static <R extends Record> R newRecordInternal(Class<R> clazz, List values) {
        List<Class<?>> types = new ArrayList<>();
        RecordComponent[] recordComponents = clazz.getRecordComponents();
        for (int i = 0; i < recordComponents.length; i++) {
            RecordComponent component = recordComponents[i];
            if (component.getType().isRecord() && !(values.get(i) instanceof Record)) {
                List nestedValues = new ArrayList<>(
                        values.subList(i, i + component.getType().getRecordComponents().length));
                R value = newRecordInternal((Class<? extends R>) component.getType(), nestedValues);
                values.add(i, value);
                values.removeAll(nestedValues);
            }
            types.add(component.getType());
        }
        try {
            Constructor<? extends Record> canonical = clazz.getDeclaredConstructor(types.toArray(Class[]::new));
            var result = (R) canonical.newInstance(values.toArray(Object[]::new));
            return result;
        } catch (ReflectiveOperationException e) {
            throw new RuntimeException("New Record creation failed: " + e, e);
        }
    }

    public void buildRecordPKFromRow(Object domainObject, AbstractRecord databaseRow) {
        EntityManagerFactory entityManagerFactory = Persistence.createEntityManagerFactory(PU_NAME);
        Session session = ((JpaEntityManager) entityManagerFactory.createEntityManager()).getActiveSession();
        RecordPK source = new RecordPK(ID_INT, ID_STRING);
        RecordPK clone = (RecordPK) session.copy(source);
        assertEquals(source, clone);
    }

    public ArrayRecord prepareRecordPKData(AbstractSession session) {
        Vector<DatabaseField> EntityFields = new Vector<>(session.getDescriptor(RecordEntity.class).getAllFields());
        DatabaseField[] recordEntityFieldsArray = EntityFields.toArray(new DatabaseField[0]);
        Object[] values = new Object[] { Integer.valueOf(1), "a", "abcde", "ijkl", "xyz" };
        ArrayRecord databaseRow = new ArrayRecord(EntityFields, recordEntityFieldsArray, values);
        return databaseRow;
    }

    public ArrayRecord prepareRecordAttributeData(AbstractSession session) {
        Vector<DatabaseField> EntityFields = new Vector<>(session.getDescriptor(RecordEntity.class).getAllFields());
        DatabaseField[] recordEntityFieldsArray = EntityFields.toArray(new DatabaseField[0]);
        // ID_INT ID_STRING NAME3 NAME1 NAME2
        Object[] values = new Object[] { Integer.valueOf(1), "a", "xyz", "abcde", "ijkl" };
        ArrayRecord databaseRow = new ArrayRecord(EntityFields, recordEntityFieldsArray, values);
        return databaseRow;
    }

    @Test
    public void createEmtyRecord() {
        NestedMasterRecord nestedMasterRecord = new NestedMasterRecord(0, null, null);
    }

    @Test
    public void testRecordClass() {
        Class<?> javaClass = RecordAttribute.class;
        if (javaClass != null && javaClass.isRecord()) {
            // if (javaClass instanceof java.lang.Record.class) {
            System.out.println("Is java.lang.Record");
        } else {
            System.out.println("Is not java.lang.Record");
        }
    }
}
